/*
			    Win
 */
#ifndef WIN_H
#define WIN_H

#include <gtk/gtk.h>
#include "tool_bar.h"
#include "convparm.h"


typedef struct _win_struct              win_struct;
typedef struct _win_opid_struct         win_opid_struct;


/*
 *	Operation ID Flags:
 */
typedef enum {
	WIN_OPID_ALLOW_MULTIPLE			= (1 << 0),
	WIN_OPID_NO_OP				= (1 << 1)
} win_opid_flags;


/*
 *	Operation IDs:
 */
typedef enum {

	WIN_OPID_SEPARATOR			= 10,

	WIN_OPID_CLOSE				= 100,

	WIN_OPID_NEW				= 1000,	/* Clear & add */
	WIN_OPID_ADD				= 1200,
	WIN_OPID_REMOVE				= 1201,
	WIN_OPID_SHIFT_UP			= 1220,
	WIN_OPID_SHIFT_DOWN			= 1221,

	WIN_OPID_CUT				= 1310,
	WIN_OPID_COPY				= 1311,
	WIN_OPID_PASTE				= 1312,
	WIN_OPID_DELETE				= 1313,	/* Remove */

	WIN_OPID_CLEAR_SRC			= 3120,
	WIN_OPID_CLEAR_TAR			= 3121,

	WIN_OPID_SELECT_ALL			= 1400,
	WIN_OPID_UNSELECT_ALL			= 1401,
	WIN_OPID_INVERT_SELECTION		= 1402,
	WIN_OPID_PROPERTIES			= 1490,

	WIN_OPID_SET_TAR_DIR			= 2000,
	WIN_OPID_SET_TAR_NAME_PREFIX		= 2005,
	WIN_OPID_SET_TAR_NAME_POSTFIX		= 2006,

	WIN_OPID_SET_TAR_VISUAL			= 2900,
	WIN_OPID_SET_TAR_COMPRESSION		= 2901,

	WIN_OPID_CONVERT			= 4000,
	WIN_OPID_CONVERT_SELECTED		= 4001,
	WIN_OPID_STOP				= 4100,
	WIN_OPID_REMOVE_CONVERTED		= 4200,
	WIN_OPID_RECURSIVE			= 4201,

	WIN_OPID_OPEN				= 7000,
	WIN_OPID_REFRESH			= 8000,
	WIN_OPID_PREVIEW_SRC			= 8100,
	WIN_OPID_PREVIEW_SRC_SHOW_DETAILS	= 8101,
	WIN_OPID_PREVIEW_TAR			= 8120,
	WIN_OPID_PREVIEW_TAR_SHOW_DETAILS	= 8121,
	WIN_OPID_VIEW_TOOLBAR			= 8200,
	WIN_OPID_VIEW_STATUSBAR			= 8210,

	WIN_OPID_NEW_IMGWORKS			= 9000,
	WIN_OPID_NEW_IMBR			= 9100,

	WIN_OPID_HELP_CONTENTS			= 10000,
	WIN_OPID_ABOUT				= 10500

} win_opid;


/*
 *	Win:
 */
struct _win_struct {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gboolean	processing;
	gint		busy_count,
			passive_busy_count,
			freeze_count,
			stop_count;
	gpointer	core;

	GList		*opid;

	GtkWidget	*main_vbox;

	gboolean	menubar_map_state;
	GtkWidget	*menubar_handle,
			*menubar;

	GtkWidget	*new_mi,
			*file_add_mi,
			*file_remove_mi,
			*close_mi,

			*add_mi,
			*remove_mi,
			*shift_up_mi,
			*shift_down_mi,
			*cut_mi,
			*copy_mi,
			*paste_mi,
			*delete_mi,
			*clear_src_mi,
			*clear_tar_mi,
			*select_all_mi,
			*unselect_all_mi,
			*invert_selection_mi,
			*properties_mi,

			*open_mi,
			*refresh_mi,
			*preview_src_micheck,
			*preview_src_show_details_micheck,
			*preview_tar_micheck,
			*preview_tar_show_details_micheck,
			*view_tool_bar_micheck,
			*view_statusbar_micheck,

			*set_tar_dir_mi,
			*set_tar_name_prefix_mi,
			*set_tar_name_postfix_mi,

			*convert_mi,
			*convert_selected_mi,
			*stop_mi,
			*remove_converted_micheck,
			*recursive_micheck,

			*new_imgworks_mi,
			*new_imbr_mi,

			*help_contents_mi,
			*about_mi;

	gboolean	tool_bar_map_state;
	GtkWidget	*tool_bar_handle;
	GtkWidget	*tool_bar;

	GtkWidget	*lists_paned1,
			*lists_paned2,
			*src_clist,
			*src_preview,
			*tar_clist,
			*tar_preview;

	GtkWidget	*focus_clist;		/* GtkCList in focus, do not
						 * destroy */

	GList		*convparm_list;

	gboolean	statusbar_map_state;
	GtkWidget	*statusbar;
	GtkWidget	*statusbar_progress,
			*statusbar_label;
	gfloat		statusbar_progress_last;

	GtkWidget	*src_list_menu,		/* Right click menu */
			*src_list_convert_mi,
			*src_list_convert_selected_mi,
			*src_list_open_mi,
			*src_list_new_imbr_mi,
			*src_list_add_mi,
			*src_list_remove_mi,
			*src_list_shift_up_mi,
			*src_list_shift_down_mi,
			*src_list_cut_mi,
			*src_list_copy_mi,
			*src_list_paste_mi,
			*src_list_delete_mi,
			*src_list_clear_mi,
			*src_list_select_all_mi,
			*src_list_unselect_all_mi,
			*src_list_invert_selection_mi,
			*src_list_properties_mi;

	GtkWidget	*tar_list_menu,
			*tar_list_open_mi,
			*tar_list_new_imbr_mi,
			*tar_list_cut_mi,
			*tar_list_copy_mi,
			*tar_list_paste_mi,
			*tar_list_delete_mi,
			*tar_list_clear_mi,
			*tar_list_select_all_mi,
			*tar_list_unselect_all_mi,
			*tar_list_invert_selection_mi,
			*tar_list_properties_mi;

	GtkWidget	*tar_dir_entry,
			*tar_dir_browse_btn,
			*tar_dir_clear_btn,
			*convert_btn,
			*imbr_btn,
			*anim_icon;

	gboolean	preview_src,
			preview_src_show_details,
			preview_tar,
			preview_tar_show_details;

	gchar		*tar_dir,
			*tar_name_prefix,
			*tar_name_postfix;

	gboolean	remove_converted,
			recursive;

};
#define WIN(p)		((win_struct *)(p))


/*
 *      Win Operation ID:
 */
struct _win_opid_struct {

	win_opid	id;
	win_opid_flags	flags;
	win_struct	*win;

	ToolBarItemType	item_type;

	gchar		*button_name,
			*menu_name,
			*tooltip;

	guint		accel_key,
			accel_mods;

	guint8		**button_icon_data,	/* Shared */
			**menu_icon_data;	/* Shared */

	void (*func_cb)(
		GtkWidget *,		/* GtkWidget */
		gpointer		/* Win */
	);
	gint (*enter_func_cb)(
		GtkWidget *,		/* GtkWidget */
		GdkEventCrossing *,	/* GdkEventCrossing */
		gpointer		/* OPID */
	);
	gint (*leave_func_cb)(
		GtkWidget *,		/* GtkWidget */
		GdkEventCrossing *,	/* GdkEventCrossing */
		gpointer		/* OPID */
	);

};
#define WIN_OPID(p)		((win_opid_struct *)(p))


/* Operation IDs */
extern win_opid_struct *WinOPIDNew(
	const win_opid id,
	win_struct *win
);
extern void WinOPIDDelete(win_opid_struct *opid);
extern GList *WinOPIDListNew(win_struct *win);
extern void WinOPIDListDelete(GList *glist);
extern win_opid_struct *WinOPIDListFind(GList *glist, win_opid id);

/* Target Directory */
extern void WinSetTarDir(win_struct *win, const gchar *path);
extern gchar *WinGetTarDir(win_struct *win);

/* Remove Converted */
void WinSetRemoveConverted(win_struct *win, gboolean remove_converted);
gboolean WinGetRemoveConverted(win_struct *win);

/* Recursive */
void WinSetRecursive(win_struct *win, gboolean recursive);
gboolean WinGetRecursive(win_struct *win);


/* Win */
extern win_struct *WinNew(gpointer core_ptr);
extern void WinSavePositions(win_struct *win);
extern void WinUpdate(win_struct *win);
extern void WinSetBusy(win_struct *win, gboolean busy);
extern void WinSetPassiveBusy(win_struct *win, gboolean busy);
extern gboolean WinToolBarIsShown(win_struct *win);
extern void WinToolBarSetShow(win_struct *win, gboolean show);
extern void WinToolBarUpdateItems(win_struct *win);
extern gboolean WinStatusBarIsShown(win_struct *win);
extern void WinStatusBarSetShow(win_struct *win, gboolean show);
extern void WinStatusProgress(
	win_struct *win, gfloat val, gboolean allow_gtk_iteration
);
extern void WinStatusMessage(
	win_struct *win, const gchar *mesg, gboolean allow_gtk_iteration
);
extern gboolean WinIsMapped(win_struct *win);
extern void WinMap(win_struct *win);
extern void WinUnmap(win_struct *win);
extern void WinDelete(win_struct *win);


#endif	/* WIN_H */
