#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "winlistctrl.h"
#include "cfg.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"


gint WinCListKeyEventCB(
	GtkCList *clist, GdkEventKey *key, gpointer data
);
gint WinCListButtonEventCB(
	GtkCList *clist, GdkEventButton *button, gpointer data
);
gint WinCListMotionEventCB(
	GtkCList *clist, GdkEventMotion *motion, gpointer data
);

static gint WinScrolledWindowEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
GtkWidget *WinScrolledWindowNew(
	core_struct *core,
	GtkPolicyType hscrollbar_policy,
	GtkPolicyType vscrollbar_policy,
	GtkWidget **event_box_rtn,
	GtkWidget **vbox_rtn
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define WIN_SCROLLED_WINDOW_LIST_CTRL_DATA	\
	"win_scrolled_window_list_ctrl"

static gboolean		do_scroll;
static gint		pointer_x_last,
			pointer_y_last;


/*
 *	GtkCList "key_press_event" or "key_release_event" signal
 *	callback.
 */
gint WinCListKeyEventCB(
	GtkCList *clist, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	guint state;
	GtkSelectionMode selection_mode;
	core_struct *core = CORE(data);
	if((clist == NULL) || (key == NULL) || (core == NULL))
	    return(status);

#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(clist),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}

#define DO_ADJ_CLAMP_EMIT	{		\
 if(adj->value > (adj->upper - adj->page_size))	\
  adj->value = adj->upper - adj->page_size;	\
 if(adj->value < adj->lower)			\
  adj->value = adj->lower;			\
 gtk_signal_emit_by_name(			\
  GTK_OBJECT(adj), "value_changed"		\
 );						\
}

#define DO_CLIST_FOCUS_ROW_CLAMP	{	\
 if(clist->focus_row >= clist->rows)		\
  clist->focus_row = clist->rows - 1;		\
 if(clist->focus_row < 0)			\
  clist->focus_row = 0;				\
}

	selection_mode = clist->selection_mode;
	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	state = key->state;

	switch(key->keyval)
	{
	  case GDK_Up:
	  case GDK_KP_Up:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Get adjustment and scroll up */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);
		    adj->value -= adj->step_increment;
		    DO_ADJ_CLAMP_EMIT
		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		if(press)
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row--;
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    0.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gtk_clist_select_row(
				clist, prev_focus_row, 0
			    );
			    gtk_clist_select_row(
				clist, clist->focus_row, 0
			    );
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Down:
	  case GDK_KP_Down:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Get adjustment and scroll down */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);
		    adj->value += adj->step_increment;
		    DO_ADJ_CLAMP_EMIT
		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		GtkAdjustment *adj = clist->hadjustment;
		if((adj != NULL) && press)
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row++;
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    1.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gtk_clist_select_row(
				clist, prev_focus_row, 0
			    );
			    gtk_clist_select_row(
				clist, clist->focus_row, 0
			    );
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Left:
	  case GDK_KP_Left:
	    if(TRUE)
	    {
		/* Get adjustment and scroll left */
		GtkAdjustment *adj = clist->hadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    if(state & GDK_CONTROL_MASK)
			adj->value -= adj->step_increment * 4;
		    else if(state & GDK_SHIFT_MASK)
			adj->value = adj->lower;
		    else
			adj->value -= adj->step_increment;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Right:
	  case GDK_KP_Right:
	    if(TRUE)
	    {
		/* Get adjustment and scroll right */
		GtkAdjustment *adj = clist->hadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    if(state & GDK_CONTROL_MASK)
			adj->value += adj->step_increment * 4;
		    else if(state & GDK_SHIFT_MASK)
			adj->value = adj->upper - adj->page_size;
		    else
			adj->value += adj->step_increment;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Page_Up:
	  case GDK_KP_Page_Up:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Scroll up one page step */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    adj->value -= adj->page_increment;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		/* Move focus row up one page step */
		gint row_height = (GTK_CLIST_ROW_HEIGHT_SET(clist) ?
		    clist->row_height : WIN_LIST_ROW_SPACING) + 1;
		const GtkAdjustment *adj = clist->vadjustment;
		if(press && (row_height > 0) && (adj != NULL))
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row -= (gint)MAX(
			(adj->page_increment / row_height), 1.0f
		    );
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    0.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gint i;
			    for(i = MAX(prev_focus_row, clist->focus_row);
				i >= clist->focus_row;
				i--
			    )
				gtk_clist_select_row(
				    clist, i, 0
				);
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Page_Down:
	  case GDK_KP_Page_Down:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Scroll down one page step */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    adj->value += adj->page_increment;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		/* Move focus row down one page step */
		gint row_height = (GTK_CLIST_ROW_HEIGHT_SET(clist) ?
		    clist->row_height : WIN_LIST_ROW_SPACING) + 1;
		const GtkAdjustment *adj = clist->vadjustment;
		if(press && (row_height > 0) && (adj != NULL))
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row += (gint)MAX(
			(adj->page_increment / row_height), 1.0f
		    );
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    1.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gint i;
			    for(i = MIN(prev_focus_row, clist->focus_row);
				i <= clist->focus_row;
				i++
			    )
				gtk_clist_select_row(
				    clist, i, 0
				);
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_Home:
	  case GDK_KP_Home:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Scroll to top */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    adj->value = adj->lower;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		/* Move focus row to top */
		if(press)
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row = 0;
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    0.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gint i;
			    for(i = MAX(prev_focus_row, clist->focus_row);
				i >= clist->focus_row;
				i--
			    )
				gtk_clist_select_row(
				    clist, i, 0
				);
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;

	  case GDK_End:
	  case GDK_KP_End:
	    if(state & GDK_CONTROL_MASK)
	    {
		/* Scroll to bottom */
		GtkAdjustment *adj = clist->vadjustment;
		if((adj != NULL) && press)
		{
		    gtk_clist_freeze(clist);

		    adj->value = adj->upper - adj->page_size;
		    DO_ADJ_CLAMP_EMIT

		    gtk_clist_thaw(clist);
		}
	    }
	    else
	    {
		/* Move focus row to top */
		if(press)
		{
		    gint prev_focus_row = clist->focus_row;

		    gtk_clist_freeze(clist);

		    clist->focus_row = clist->rows - 1;
		    DO_CLIST_FOCUS_ROW_CLAMP

		    if(gtk_clist_row_is_visible(
			clist, clist->focus_row) != GTK_VISIBILITY_FULL
		    )
			gtk_clist_moveto(
			    clist,
			    clist->focus_row, -1,	/* Row, column */
			    1.0f, 0.0f			/* Row, column */
			);

		    /* Select rows in between? */
		    if((selection_mode == GTK_SELECTION_MULTIPLE) ||
		       (selection_mode == GTK_SELECTION_EXTENDED)
		    )
		    {
			if(state & GDK_SHIFT_MASK)
			{
			    gint i;
			    for(i = MIN(prev_focus_row, clist->focus_row);
				i <= clist->focus_row;
				i++
			    )
				gtk_clist_select_row(
				    clist, i, 0
				);
			}
		    }

		    gtk_clist_thaw(clist);
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;
	}

	return(status);
#undef DO_CLIST_FOCUS_ROW_CLAMP
#undef DO_ADJ_CLAMP_EMIT
#undef SIGNAL_EMIT_STOP
}

/*
 *	GtkCList "button_press_event" or "button_release_event" signal
 *	callback.
 */
gint WinCListButtonEventCB(
	GtkCList *clist, GdkEventButton *button, gpointer data
)
{
	gint status = FALSE;
	core_struct *core = CORE(data);
	if((clist == NULL) || (button == NULL) || (core == NULL))
	    return(status);

#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(clist),				\
  (button->type == GDK_BUTTON_PRESS) ?		\
   "button_press_event" : "button_release_event"\
 );						\
}

	switch((gint)button->type)
	{
	  case GDK_BUTTON_PRESS:
	    switch(button->button)
	    {
	      case GDK_BUTTON2:
#ifdef HAVE_LIBENDEAVOUR2
		if(edv_get_i(
		    core->edv_ctx,
		    EDV_CFG_PARM_LISTS_POINTER_OP_BUTTON2
		) == EDV_LISTS_POINTER_OP_BUTTON2_SCROLL_XY)
#else
		if(TRUE)
#endif
		{
		    gint x, y;
		    GdkModifierType mask;
		    GdkCursor *cur = core->translate_cur;
		    gdk_window_get_pointer(
			button->window, &x, &y, &mask
		    );
		    gdk_pointer_grab(
			button->window,
			FALSE,
			GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_POINTER_MOTION_MASK |
			    GDK_POINTER_MOTION_HINT_MASK,
			NULL,
			cur,
			button->time
		    );
		    do_scroll = TRUE;
		    pointer_x_last = x;
		    pointer_y_last = y;
		    SIGNAL_EMIT_STOP
		    status = TRUE;
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    switch(button->button)
	    {
	      case GDK_BUTTON2:
#ifdef HAVE_LIBENDEAVOUR2
		if(edv_get_i(
		    core->edv_ctx,
		    EDV_CFG_PARM_LISTS_POINTER_OP_BUTTON2
		) == EDV_LISTS_POINTER_OP_BUTTON2_SCROLL_XY)
#else
		if(TRUE)
#endif
		{
		    gdk_pointer_ungrab(button->time);

		    do_scroll = FALSE;
		    pointer_x_last = 0;
		    pointer_y_last = 0;

		    SIGNAL_EMIT_STOP
		    status = TRUE;
		}
		break;
	    }
	    break;
	}

	return(status);
#undef SIGNAL_EMIT_STOP
}

/*
 *	GtkCList "motion_notify_event" signal callback.
 */
gint WinCListMotionEventCB(
	GtkCList *clist, GdkEventMotion *motion, gpointer data
)
{
	gint status = FALSE;
	core_struct *core = CORE(data);
	if((clist == NULL) || (motion == NULL) || (core == NULL))
	    return(status);

#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(clist), "motion_notify_event"	\
 );						\
}

#define DO_SCROLL(dx,dy)	{		\
 GtkAdjustment *adj = clist->hadjustment;	\
 if((adj != NULL) && ((dx) != 0)) {		\
  if((adj->upper - adj->page_size) > adj->lower) { \
   adj->value = CLIP(				\
    adj->value - (dx),				\
    adj->lower, adj->upper - adj->page_size	\
   );						\
   gtk_signal_emit_by_name(			\
    GTK_OBJECT(adj), "value_changed"		\
   );						\
  }						\
 }						\
 adj = clist->vadjustment;			\
 if((adj != NULL) && ((dy) != 0)) {		\
  if((adj->upper - adj->page_size) > adj->lower) { \
   adj->value = CLIP(				\
    adj->value - (dy),				\
    adj->lower, adj->upper - adj->page_size	\
   );						\
   gtk_signal_emit_by_name(			\
    GTK_OBJECT(adj), "value_changed"		\
   );						\
  }						\
 }						\
}

	switch((gint)motion->type)
	{
	  case GDK_MOTION_NOTIFY:
	    if(do_scroll)
	    {
		gint dx, dy;
		if(motion->is_hint)
		{
		    gint x, y;
		    GdkModifierType mask;
		    gdk_window_get_pointer(
			motion->window, &x, &y, &mask
		    );
		    dx = x - pointer_x_last;
		    dy = y - pointer_y_last;
		    pointer_x_last = x;
		    pointer_y_last = y;
		}
		else
		{
		    dx = (gint)(motion->x - pointer_x_last);
		    dy = (gint)(motion->y - pointer_y_last);
		    pointer_x_last = (gint)motion->x;
		    pointer_y_last = (gint)motion->y;
		}
		DO_SCROLL(dx, dy);
		SIGNAL_EMIT_STOP
		status = TRUE;
	    }
	    break;
	}

	return(status);
#undef DO_SCROLL
#undef SIGNAL_EMIT_STOP
}


/*
 *	Scrolled window GtkEventBox event signal callback.
 */
static gint WinScrolledWindowEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	GList *glist;
	GdkEventFocus *focus;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	GtkScrolledWindow *sw;
	core_struct *core;
	if((widget == NULL) || (event == NULL))
	    return(status);

	glist = (GList *)data;
	sw = (glist != NULL) ? (GtkScrolledWindow *)glist->data : NULL;
	glist = g_list_next(glist);
	core = (glist != NULL) ? CORE(glist->data) : NULL;
	if((sw == NULL) || (core == NULL))
	    return(status);

#define ADJ_SET_EMIT(adj,v)	{		\
 gfloat v2 = (v);				\
 if(v2 > ((adj)->upper - (adj)->page_size))	\
  v2 = (adj)->upper - (adj)->page_size;		\
 if(v2 < (adj)->lower)				\
  v2 = (adj)->lower;				\
 gtk_adjustment_set_value((adj), v2);		\
}
#define ADJ_SET_DELTA_EMIT(adj,d)	{	\
 ADJ_SET_EMIT((adj), (adj)->value + (d));	\
}

	switch((gint)event->type)
	{
	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->window == widget->window)
	    {
		if(focus->in && !GTK_WIDGET_HAS_FOCUS(widget))
		{
		    GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
		    gtk_widget_queue_draw(widget);
		}
		else if(!focus->in && GTK_WIDGET_HAS_FOCUS(widget))
		{
		    GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
		    gtk_widget_queue_draw(widget);
		}
		status = TRUE;
	    }
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}
	    switch(key->keyval)
	    {
	      case GDK_Up:
	      case GDK_KP_Up:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, -adj->step_increment);
		}
		SIGNAL_EMIT_STOP
		status = TRUE;
		break;
	      case GDK_Down:
	      case GDK_KP_Down:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, adj->step_increment);
		}
		SIGNAL_EMIT_STOP
		status = TRUE;
		break;
	      case GDK_Left:
	      case GDK_KP_Left:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_hadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, -adj->step_increment);
		}
		SIGNAL_EMIT_STOP 
		status = TRUE;
		break;
	      case GDK_Right:
	      case GDK_KP_Right:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_hadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, adj->step_increment);
		}
		SIGNAL_EMIT_STOP 
		status = TRUE;
		break;
	      case '-':
	      case 'b':
	      case 'p':
	      case GDK_Page_Up:
	      case GDK_KP_Page_Up:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, -adj->page_increment);
		}
		status = TRUE;
		break;
	      case 'n':
	      case GDK_space:
	      case GDK_Page_Down:
	      case GDK_KP_Page_Down:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_DELTA_EMIT(adj, adj->page_increment);
		}
		status = TRUE;
		break;
	      case GDK_Home:
	      case GDK_KP_Home: 
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_EMIT(adj, adj->lower);
		}
		status = TRUE;
		break;
	      case GDK_End:
	      case GDK_KP_End:
		if(press)
		{
		    GtkAdjustment *adj = gtk_scrolled_window_get_vadjustment(sw);
		    ADJ_SET_EMIT(adj, adj->upper - adj->page_size);
		}
		status = TRUE;
		break;
	    }
#undef SIGNAL_EMIT_STOP  
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(button->window == widget->window)
	    {
		if(!GTK_WIDGET_HAS_FOCUS(widget))
		    gtk_widget_grab_focus(widget);
	    }
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
	      case GDK_BUTTON2:
		if(button->window == widget->window)
		{
		    gdk_pointer_grab(
			button->window,
			FALSE,
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK |
			GDK_POINTER_MOTION_HINT_MASK,
			NULL,
			core->translate_cur,
			button->time
		    );
		    do_scroll = TRUE;
		    pointer_x_last = (gint)button->x;
		    pointer_y_last = (gint)button->y;
		    status = TRUE;
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
	      case GDK_BUTTON2:
		if(do_scroll)
		{
		    if(gdk_pointer_is_grabbed())
			gdk_pointer_ungrab(button->time);
		    do_scroll = FALSE;
		    pointer_x_last = 0;
		    pointer_y_last = 0;
		    status = TRUE;
		}
		break;
	    }
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(do_scroll)
	    {
		gint dx, dy;
		GtkAdjustment *hadj = gtk_scrolled_window_get_hadjustment(sw),
			      *vadj = gtk_scrolled_window_get_vadjustment(sw);
		if(motion->is_hint)  
		{
		    gint x, y;
		    GdkModifierType mask;
		    gdk_window_get_pointer(
			motion->window, &x, &y, &mask
		    );
		    dx = x - pointer_x_last;
		    dy = y - pointer_y_last;
		}
		else
		{
		    dx = (gint)(motion->x - pointer_x_last);
		    dy = (gint)(motion->y - pointer_y_last);
		}
#define DO_SCROLL(adj,d)	{			\
 if((d) != 0) {						\
  if(((adj)->upper - (adj)->page_size) > (adj)->lower) {\
   gtk_adjustment_set_value(				\
    adj,						\
    CLIP(						\
     (adj)->value - (d),				\
     (adj)->lower, (adj)->upper - (adj)->page_size	\
    )							\
   );							\
} } }
		DO_SCROLL(hadj, dx);
		DO_SCROLL(vadj, dy);
#undef DO_SCROLL
		status = TRUE;
	    }
	    break;
	}

#undef ADJ_SET_DELTA_EMIT
#undef ADJ_SET_EMIT

	return(status);
}

/* 
 *	Creates a new GtkScrolledWindow.
 *
 *	It will be set up for scrolling.
 */
GtkWidget *WinScrolledWindowNew(
	core_struct *core,
	GtkPolicyType hscrollbar_policy,
	GtkPolicyType vscrollbar_policy,
	GtkWidget **event_box_rtn,
	GtkWidget **vbox_rtn
)
{
	GList *glist, *data_list;
	GtkWidget *parent, *w = gtk_scrolled_window_new(NULL, NULL);
	GtkScrolledWindow *sw = GTK_SCROLLED_WINDOW(w);
	gtk_scrolled_window_set_policy(
	    sw, hscrollbar_policy, vscrollbar_policy
	);
	parent = w;

	glist = NULL;
	glist = g_list_append(glist, sw);
	glist = g_list_append(glist, core);
	data_list = glist;
	gtk_object_set_data_full(
	    GTK_OBJECT(w),
	    WIN_SCROLLED_WINDOW_LIST_CTRL_DATA,
	    data_list,
	    (GtkDestroyNotify)g_list_free
	);

	w = gtk_event_box_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
	gtk_widget_add_events(
	    w,
	    GDK_FOCUS_CHANGE_MASK |
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(WinScrolledWindowEventCB), data_list
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(WinScrolledWindowEventCB), data_list
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(WinScrolledWindowEventCB), data_list
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(WinScrolledWindowEventCB), data_list
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(WinScrolledWindowEventCB), data_list
	);
	gtk_scrolled_window_add_with_viewport(sw, w);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, 0);
	if(vbox_rtn != NULL)
	    *vbox_rtn = w;
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);

	return(GTK_WIDGET(sw));
}
